'''
Handles the rough edges of AWS SNS notifications for modular use of notifications.
'''

import boto3
import traceback as tb
from kizano import getLogger, getConfig

log = getLogger(__name__)
sns = boto3.client('sns')

class NotifyHelper(object):
    def __init__(self):
        self.config = getConfig()
        if 'aws' in self.config and 'sns' in self.config['aws']:
            self.notify_topic = self.config['aws']['sns']['notify_topic']
            self.error_topic = self.config['aws']['sns']['error_topic']
        else:
            topics = sns.list_topics()
            if 'Topics' in topics:
                log.debug('Topics: %s' % topics['Topics'])
                # Set self.notify_topic to the config `.aws.sns.notify_topic` or a topic with "INFO" in the name if config not set.
                self.notify_topic = next((x['TopicArn'] for x in topics['Topics'] if 'INFO' in x['TopicArn'].upper()))
                # Same with `self.error_topic` but with "ERROR" in the name.
                self.error_topic = next((x['TopicArn'] for x in topics['Topics'] if 'ERROR' in x['TopicArn'].upper()))
            else:
                log.error('No topics found!')
                self.notify_topic = ''
                self.error_topic = ''

    def info(self, subject: str, message: str) -> None:
        '''
        Send an INFO message to the configured SNS topic.
        '''
        if self.notify_topic:
            return sns.publish(
                TopicArn=self.notify_topic,
                Subject=subject,
                Message=message,
            )
        else:
            log.warning('No notify_topic set in config!')

    def error(self, subject: str, message: str) -> None:
        '''
        Send an ERROR message to the configured SNS topic.
        '''
        if self.error_topic:
            return sns.publish(
                TopicArn=self.error_topic,
                Subject=subject,
                Message=f'{message}\nTraceback:\n{tb.format_exc()}',
            )
        else:
            log.warning('No error_topic set in config!')



'''
\x1b[31mModule-Level Documentation!\x1b[0m
'''
import os
import dateparser
import pymongo
import json
from datetime import datetime

from kizano import getConfig, getLogger
from smartmetertx.api import MeterReader
from smartmetertx.notify import NotifyHelper
from smartmetertx.utils import getMongoConnection

log = getLogger(__name__)
HOME = os.getenv('HOME', '')
SMTX_FROM   = dateparser.parse(os.environ.get('SMTX_FROM', 'day before yesterday'))
SMTX_TO     = dateparser.parse(os.environ.get('SMTX_TO', 'today'))

class Smtx2Mongo(object):
    '''
    SmartMeterTexas -> MongoDB
    Model object to take the records we get from https://smartmetertexas.com and insert them into
    a mongodb we control for data preservation and other analytics on our electric usage data we
    would want to undertake.
    '''

    def __init__(self):
        self.config = getConfig()
        self.notify = NotifyHelper()
        self.mongo = getMongoConnection(self.config)
        self.db = self.mongo.get_database(self.config['mongo'].get('dbname', 'smartmetertx'))
        self.dailyReads = self.db.dailyReads
        self.interval15minReads = self.db.interval15minReads
        self.getSMTX()
        self.ensureIndexes()

    def close(self):
        if self.mongo:
            self.mongo.close()
            self.mongo = None

    def ensureIndexes(self):
        self.dailyReads.create_index(
            [('revisionDate', 1)],
            background=True,
        )
        self.dailyReads.create_index(
            [('readDate', 1)],
            background=True,
            unique=True,
        )
        self.interval15minReads.create_index(
            [('readingDate', 1)],
            background=True,
            unique=True,
        )
        self.interval15minReads.create_index(
            [('revisionDate', 1)],
            background=True,
            unique=True,
        )
        return self

    def getSMTX(self):
        log.info('Connecting to SmartMeterTX...')
        self.smtx = MeterReader()
        log.info('Success!')
        return self.smtx

    def getDailyReads(self):
        # Get the meter reads and print the date in the format their API expects.
        log.info('Getting daily reads from SmartMeterTX API...')
        reads = self.smtx.get_daily_read(self.config['smartmetertx']['esiid'], SMTX_FROM.strftime('%m/%d/%Y'), SMTX_TO.strftime('%m/%d/%Y'))
        log.info('Acquired %d meter reads!' % len(reads['registeredReads']))
        return reads

    def get15minReads(self):
        log.info('Getting 15min reads from SmartMeterTX API...')
        reads = self.smtx.get_15min_reads(self.config['smartmetertx']['esiid'], SMTX_FROM.strftime('%m/%d/%Y'), SMTX_TO.strftime('%m/%d/%Y'))
        log.info('Acquired %d meter reads!' % len(reads['energyData']))
        return reads

    def typecastDailyReads(self, dailyData: list[dict]) -> list[dict]:
        '''
        Convert strings to proper data types to store in DB.
        Feb-2024 update data structure:

        {
            "trans_id": "00000000000000000",
            "esiid": "0000000000000000000",
            "registeredReads": [{
                "readDate": "01/01/2024",
                "revisionDate": "01/02/2024 00:00:00",
                "startReading": "0000.000",
                "endReading": "0000.000",
                "energyDataKwh": "00.000"
            }]
        }
        '''
        results = []
        for meterRead in dailyData:
            meterRead['readDate'] = datetime.strptime(meterRead['readDate'], '%m/%d/%Y')
            meterRead['revisionDate'] = datetime.strptime(meterRead['revisionDate'], '%m/%d/%Y %H:%M:%S')
            meterRead['startReading'] = float(meterRead['startReading'])
            meterRead['endReading'] = float(meterRead['endReading'])
            meterRead['energyDataKwh'] = float(meterRead['energyDataKwh'])
            results.append(meterRead)
        log.debug(json.dumps(dailyData, indent=2, default=str))
        return results

    def typecast15minReads(self, interval15Data: list[dict]) -> list[dict]:
        '''
        Convert strings to proper date types to store in DB.
        Sample data structure:

        {
            "trans_id": "whateveryouputfortxnid",
            "esiid": "00000000000000000",
            "energyData": [
                {
                    "DT": "09/01/2025",
                    "RevTS": "09/02/2025 06:21:18",
                    "RT": "C",
                    "RD": "0.155-A,0.474-A,0.282-A,0.304-A,0.286-A,0.293-A,0.294-A,0.310-A,,,,,0.394-A"
                }
            ]
        }

        "RD" field contains 100 comma-separated readings, but the 4 empty ones removed makes 96.
        It accounts for 2 hours between the break, so ... timezone offset?
        '''
        result = []
        log.debug(json.dumps(interval15Data, indent=2, default=str))
        # Re-map the keys to make them more descriptive.
        for energyData in interval15Data:
            remapped = {
                'readingDate': dateparser.parse(energyData['DT']),
                'revisionDate': dateparser.parse(energyData['RevTS']),
                'readingType': energyData['RT'],
                'readingData': [],
            }
            for readingData in energyData['RD'].split(','):
                if not readingData: continue
                if readingData == '-':
                    remapped['readingData'].append( (0.0, 'X') )
                    continue
                reading, readingType = readingData.split('-')
                reading = float(reading)
                remapped['readingData'].append( (reading, readingType) )
            result.append(remapped)
        return result

    def insertDailyData(self, dailyData):
        log.info('Inserting %d reads into the DB.' % len(dailyData))
        try:
            insertResult = self.dailyReads.insert_many(dailyData, ordered=False)
            log.debug(insertResult)
        except pymongo.errors.BulkWriteError as e:
            errs = list(filter( lambda x: x['code'] != 11000, e.details['writeErrors'] ))
            if errs:
                log.error('Failed to insert daily reads: %s' % errs)
                self.notify.error('SmartMeterTX to MongoDB Exception', f'Failed to insert daily reads into MongoDB:\n{errs}')
        log.info('Complete!')

    def insert15minData(self, interval15minData):
        log.info('Inserting %d 15min reads into the DB.' % len(interval15minData))
        try:
            insertResult = self.interval15minReads.insert_many(interval15minData, ordered=False)
            log.debug(insertResult)
        except pymongo.errors.BulkWriteError as e:
            errs = list(filter( lambda x: x['code'] != 11000, e.details['writeErrors'] ))
            if errs:
                log.error('Failed to insert 15min reads: %s' % errs)
                self.notify.error('SmartMeterTX to MongoDB Exception', f'Failed to insert 15min reads into MongoDB:\n{errs}')
        log.info('Complete!')

def main() -> int:
    log.info('Gathering records from %s to %s' % ( SMTX_FROM.strftime('%F/%R'), SMTX_TO.strftime('%F/%R') ) )
    smtx2mongo = Smtx2Mongo()
    try:
        dailyReads = smtx2mongo.getDailyReads()
        dailyData = smtx2mongo.typecastDailyReads(dailyReads['registeredReads'])
        if dailyData:
            smtx2mongo.insertDailyData(dailyData)
        else:
            log.warning('No daily reads inserted!')
    except Exception as e:
        errmsg = f'Global exception trying to insert daily reads into MongoDB:\n{e}\n'
        log.error(errmsg)
        smtx2mongo.notify.error('SmartMeterTX to MongoDB Exception', errmsg)
        return 1

    try:
        interval15minReads = smtx2mongo.get15minReads()
        interval15minData = smtx2mongo.typecast15minReads(interval15minReads['energyData'])
        if interval15minData:
            smtx2mongo.insert15minData(interval15minData)
        else:
            log.warning('No 15min reads inserted!')
    except Exception as e:
        errmsg = f'Global exception trying to insert 15min reads into MongoDB:\n{e}\n'
        log.error(errmsg)
        smtx2mongo.notify.error('SmartMeterTX to MongoDB Exception', errmsg)
        return 1

    smtx2mongo.close()
    return 0




'''
SMTX Meter Report

This script is used to generate a report of the daily meter reads from the SmartMeterTX API based
on what is stored in the MongoDB. Sends a report via AWS SNS to the configured topic.
'''

import os
import dateparser

from kizano import getConfig, getLogger
from smartmetertx.utils import getMongoConnection
from smartmetertx.notify import NotifyHelper

log = getLogger(__name__)
HOME = os.getenv('HOME', '')
SMTX_FROM   = dateparser.parse(os.environ.get('SMTX_FROM', 'day before yesterday'))
SMTX_TO     = dateparser.parse(os.environ.get('SMTX_TO', 'today'))

class SmartMeterTxMeterReport(object):

    def __init__(self):
        self.config = getConfig()
        self.notify = NotifyHelper()
        self.mongo = getMongoConnection(self.config)
        self.db = self.mongo.get_database(self.config['mongo'].get('dbname', 'smartmetertx'))
        self.collection = self.db.dailyReads

    def close(self):
        if self.mongo:
            self.mongo.close()
            self.mongo = None

    def getDailyReads(self):
        '''
        Get the meter reads for the date range specified in the environment variables.
        '''
        log.info('Getting daily reads from SmartMeterTX DB...')
        reads = self.collection.find({
            'readDate': {
                '$gte': SMTX_FROM,
                '$lte': SMTX_TO
            }
        }).sort('readDate', 1)
        if not reads:
            log.error('No records found for the date range specified.')
            return None
        result = []
        for reading in reads:
            reading['readDate'] = reading['readDate'].strftime('%F')
            reading['energyDataKwh'] = float(reading['energyDataKwh'])
            result.append(reading)
        return result

    def sendReport(self, reads):
        '''
        Send the report via AWS SNS.
        '''
        log.info('Sending report via AWS SNS...')
        report = '''
Daily Reads Report
==================

Date Range: %s to %s
    
''' % ( SMTX_FROM.strftime('%F'), SMTX_TO.strftime('%F') )
        total = 0.0
        for i, read in enumerate(reads):
            report += '%(readDate)s: %(energyDataKwh)03.3f kWh    ' % read
            if (i + 1) % 7 == 0:
                report += '\n'
            total += float(read['energyDataKwh'])
        report += "\n\nTotal Energy Use: %0.3f\n" % total
        log.debug(report)
        response = self.notify.info('SmartMeterTX Daily Reads Report', report)
        log.info(f'Report size in {len(report.encode("utf-8"))} bytes with {len(reads)} records and message id {response["MessageId"]}.')

def main() -> int:
    log.info('Gathering records from %s to %s' % ( SMTX_FROM.strftime('%F'), SMTX_TO.strftime('%F') ) )
    smtxReport = SmartMeterTxMeterReport()
    reads = smtxReport.getDailyReads()
    if not reads:
        log.error('Failed to read SmartMeterTX db...')
        return 2
    smtxReport.sendReport(reads)
    return 0




import os
import requests
import gnupg
import dateparser
from datetime import datetime

from kizano import getLogger, getConfig

# BEGIN: #StackOverflow
# @Source: https://stackoverflow.com/a/16630836/2769671
# These two lines enable debugging at httplib level (requests->urllib3->http.client)
# You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
# The only thing missing will be the response.body which is not logged.
if os.getenv('DEBUG', False):
    requests_log = getLogger("requests.urllib3", 10)
    requests_log.propagate = True
# END: #StackOverflow

class MeterReaderException(Exception):
    def __init__(self, url: str, request: dict, previous: Exception):
        self.url = url
        self.request = request
        self.previous = previous
        super().__init__(f"MeterReaderException: {url}, {request}, {previous}")

class MeterReader:
    HOSTNAME = 'services.smartmetertexas.net'
    HOST = f'https://{HOSTNAME}'
    USER_AGENT = 'API Calls (python3; Linux x86_64) Track your own metrics with SmartMeterTX: https://github.com/markizano/smartmetertx'
    TIMEOUT = 30

    def __init__(self, timeout: int = 10):
        self.log = getLogger(__name__)
        self.config = getConfig()
        self.logged_in = False
        self.gpg = gnupg.GPG(gnupghome=os.path.join(os.environ['HOME'], '.gnupg'), use_agent=True)
        self.session = requests.Session()
        self.timeout = timeout
        self.lastError = None
        self.expiry = None
        self.token = None
        self.session.headers.update({
            'Authority': MeterReader.HOSTNAME,
            'Origin': MeterReader.HOST,
            'Accept': 'application/json, text/plain, */*',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/json; charset=UTF-8',
            'dnt': '1',
            'User-Agent': MeterReader.USER_AGENT,
        })
        if os.getenv('LOCAL', False):
            # I use a local proxy for just the SMTX API because they whitelist by IP.
            # Using a local SSH Dynamic Proxy allows me to source from PROD without interfering with boto3.
            self.log.info('Using local proxy...')
            self.session.proxies.update({
                'http': 'socks5://127.0.0.1:5000',
                'https': 'socks5://127.0.0.1:5000',
            })
        self.login()

    def login(self) -> str:
        '''
        Since 2025-09, SMT has updated to implement JWT.
        To authenticate, we no longer send credentials in the POST method.
        Similar to OAuth2, you post credentials to a login API endpoint and cache a credential
        received from the token generation API endpoint.
        '''
        if self.expiry and self.expiry > datetime.now():
            self.log.debug(f'Token is still valid until {self.expiry}')
            return self.token
        url = f"{MeterReader.HOST}/v2/token/"
        self.log.debug(f'Logging in to {url}...')
        json = {
            "username": self.config['smartmetertx']['user'],
            "password": self.gpg.decrypt(self.config['smartmetertx']['pass']).data.decode('utf-8')
        }
        try:
            response = self.session.post(
                url=url,
                json=json,
                timeout=self.timeout,
                # Since Feb2024 update, you need to be whitelisted and have a client cert to access the endpoint.
                # As of 2025-09, you no longer need a client cert to access the endpoint.
                #cert=(os.path.expanduser(self.config['smartmetertx']['cert_path']), os.path.expanduser(self.config['smartmetertx']['key_path'])),
                auth=(self.config['smartmetertx']['user'], self.gpg.decrypt(self.config['smartmetertx']['pass']).data.decode('utf-8'))
            )
        except Exception as ex:
            self.log.error(repr(ex))
            raise MeterReaderException(url, json, ex)
        payload = response.json()
        self.token = payload['accessToken']
        tokenType = payload['tokenType']
        self.session.headers.update({
            'Authorization': f'{tokenType} {self.token}'
        })
        self.expiry = dateparser.parse(payload['expiresAt'])
        return self.token

    def api_call(self, url: str, json: dict) -> requests.Response:
        '''
        Generic API call that can be made to the site for JSON results back.
        @param url :string: Where to send POST request.
        @param json :object: Data to send to the server.
        @return :object: JSON response back or ERROR
        '''
        self.login()
        self.log.debug(f'MeterReader.api_call(url={url}, json={json})')
        try:
            return self.session.post(
                url=url,
                json=json,
                timeout=self.timeout,
            )
        except Exception as ex:
            self.log.error(repr(ex))
            raise MeterReaderException(url, json, ex)

    def get_daily_read(self, esiid: str, start_date: str, end_date: str) -> dict:
        '''
        Gets a daily meter read.
        @param esiid :string: The ESIID to get the daily read.
        @param start_date :string: The start date to get the read in MM/DD/YYYY format.
        @param end_date :string: The end date to get the read in MM/DD/YYYY format.
        @throws Exception: If the API call fails.
        @return :object: JSON response back or False if failed.
        '''
        json = {
            "trans_id": esiid,
            "requestorID": self.config['smartmetertx']['user'].upper(),
            "requesterType": "RES",
            "startDate": start_date,
            "endDate": end_date,
            "version": "L",
            "readingType": "C",
            "esiid": [ esiid ],
            "SMTTermsandConditions": "Y"
        }
        url = f"{MeterReader.HOST}/v2/dailyreads/"
        response = self.api_call(url, json=json)
        if response.status_code != 200 or "error" in response.text.lower():
            ex = Exception(f'Failed fetching daily read! {response.text}')
            raise MeterReaderException(url, json, ex)
        return response.json()

    def get_15min_reads(self, esiid: str, start_date: str, end_date: str) -> dict|bool:
        '''
        Gets a 15 minute interval meter read.
        @param esiid :string: The ESIID to get the daily read.
        @param start_date :string: The start date to get the read in MM/DD/YYYY format.
        @param end_date :string: The end date to get the read in MM/DD/YYYY format.
        @throws Exception: If the API call fails.
        @return :object: JSON response back or False if failed.
        '''
        json = {
            "trans_id": esiid,
            "requestorID": self.config['smartmetertx']['user'].upper(),
            "requesterType": "RES",
            "startDate": start_date,
            "endDate": end_date,
            "version": "L",
            "readingType": "C",
            "esiid": [ esiid ],
            "SMTTermsandConditions": "Y"
        }
        url = f"{MeterReader.HOST}/v2/15minintervalreads/"
        response = self.api_call(url, json=json)
        if response.status_code != 200 or "error" in response.text.lower():
            ex = Exception(f'Failed fetching 15 minute interval read! {response.text}')
            raise MeterReaderException(url, json, ex)
        return response.json()

